/* MouseSync.cpp
Synchronises the mouse pointers between RISC-OS in Red Squirrel and Win32
(C) 2002 Niall Douglas
8th August 2002 (started 8th August 2002)

This file is provided under the GNU General Public Licence.
See the file gpl.txt in the archive for more information.
*/

// Include method of triggering RISC-OS code
#include "SignalVector.h"
#define _WIN32_WINNT 0x500
#include "windows.h"
#include <stdio.h>

// This is the stickyness factor
#define MOUSE_STICKYNESS 16

// Offer your API's in non-mangled form for CallWin32 to access
#ifdef __cplusplus
extern "C" {
#endif
// Expose your API's to others
#define EXPORT __declspec(dllexport)

// Some static data
static HWND parentwndh, mainwndh;
static HHOOK getmsgh, callwndproch;
static int ROvector;
static bool tracking=false;
static bool cursorhidden;
static bool IAmDisabled;

// Scaling info to convert Win32 position to RO
typedef struct ScalingInfo_t
{
	int xoffset, yoffset;		// Distance of RO screen to RS client area
	int width;
	int height;					// For inverting y as RO runs other way
	int xeig, yeig;				// Eigen factors for multipixel modes
	double RSxscale, RSyscale;	// For when RS is hardware scaling its window
} ScalingInfo;
static ScalingInfo scalinginfo;

// Structure sent to RISC-OS with mouse info
typedef struct MouseInfo_t
{
	int x,y, wheeldelta;
} MouseInfo;
static MouseInfo mouseinfo;

// Useful function for decoding a Win32 error
static void ReportError(DWORD ecode)
{
	int len;
	char msg[256];
	len=FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS, 0, ecode, 0, msg, 256, 0);
	MessageBox(NULL, msg, "Error in CallWin32", MB_OK);
}
#define ERRH(ret) if(!(ret)) ReportError(GetLastError())

// Takes a Win32 screen coordinate and makes it relative to the client area
static void ConvertMouse(MouseInfo *mi, int x, int y, int wheeldelta)
{
	POINT p;
	p.x=x; p.y=y;
	ScreenToClient(mainwndh, &p);
	mi->x=p.x; mi->y=p.y; mi->wheeldelta=wheeldelta;
}

// Called when RS switches between full screen and desktop
static void RSUpcallHandler(int op, void *data)
{
	if(0==op)
	{	// Desktop
		IAmDisabled=false;
	}
	else if(1==op)
	{	// Full screen
		IAmDisabled=true;
		CallWinPodule_SetMouseInputDisable(3);
	}
}

// Called with RS changes its window size
static void RSClientChange(HWND hwnd)
{
	RECT r;
	ERRH(GetClientRect(hwnd, &r));
	scalinginfo.xoffset=(r.right-scalinginfo.width)/2;
	scalinginfo.yoffset=(r.bottom-scalinginfo.height)/2;
	//scalinginfo.RSxscale=(double) r.right/scalinginfo.width;
	//scalinginfo.RSyscale=(double) r.bottom/scalinginfo.height;
}

// Sends the mouse info to RO
static void SendMouseInfo(MouseInfo *mi)
{
	//char temp[256];
	//sprintf(temp,"Raw x=%d,y=%d,w=%d\n", mi->x, mi->y, mi->wheeldelta);
	//OutputDebugString(temp);
	RSClientChange(mainwndh);
	mouseinfo.x=(mi->x-scalinginfo.xoffset)*2;
	mouseinfo.y=(scalinginfo.height-(mi->y-scalinginfo.yoffset))*2;
	mouseinfo.wheeldelta=mi->wheeldelta;
	// Scale mouse according to if RS is hardware scaling
	//mouseinfo.x=(int) (mouseinfo.x*scalinginfo.RSxscale);
	//mouseinfo.y=(int) (mouseinfo.y*scalinginfo.RSyscale);
	//sprintf(temp,"Sending x=%d,y=%d,w=%d\n", mouseinfo.x, mouseinfo.y, mouseinfo.wheeldelta);
	//OutputDebugString(temp);
	if(mouseinfo.x>=-MOUSE_STICKYNESS && mouseinfo.y>=-MOUSE_STICKYNESS
		&& mouseinfo.x<=(scalinginfo.width*2)+MOUSE_STICKYNESS
		&& mouseinfo.y<=(scalinginfo.height*2)+MOUSE_STICKYNESS)
	{
		if(!cursorhidden)
		{
			ShowCursor(false);
			cursorhidden=true;
		}
		CallWinPodule_SignalVector(ROvector);
	}
	else if(cursorhidden)
	{
		ShowCursor(true);
		cursorhidden=false;
	}
}

// Parent intercept
LRESULT CALLBACK WndProcIntercept(int code, WPARAM wParam, LPARAM lParam)
{
	if(HC_ACTION==code)
	{
		CWPSTRUCT *msg=(CWPSTRUCT *) lParam;
		if(msg->hwnd==parentwndh)
		{
			if(WM_SETCURSOR==msg->message /*&& (HWND) msg->wParam==mainwndh*/) // Used to start and stop mouse tracking plus hide cursor
			{
				if(!IAmDisabled)
				{
					if(LOWORD(msg->lParam)==HTCLIENT && GetForegroundWindow()==parentwndh)
					{
						tracking=true;
						CallWinPodule_SetMouseInputDisable(1);	// Disable motion
						//OutputDebugString("tracking=1\n");
					}
					else if(tracking)
					{
						tracking=false;
						if(cursorhidden)
						{
							ShowCursor(true);
							cursorhidden=false;
						}
						CallWinPodule_SetMouseInputDisable(0);	// Disable everything
						//OutputDebugString("tracking=0\n");
					}
				}
			}
		}
	}
	return CallNextHookEx(callwndproch, code, wParam, lParam);
}

// Intercepts all messages posted to Red Squirrel
LRESULT CALLBACK GetMsgIntercept(int code, WPARAM wParam, LPARAM lParam)
{
	if(HC_ACTION==code)
	{
		MSG *msg=(MSG *) lParam;
		//char temp[256];
		if(msg->hwnd==mainwndh)
		{
			bool eatit=false;
		//sprintf(temp,"wndh=0x%x, msg=0x%x, tracking=%d\n",msg->hwnd, msg->message, tracking);
		//OutputDebugString(temp);
			switch(msg->message)
			{
				case WM_MOUSEMOVE:		// Used to sync RO mouse ptr
				{
					if(tracking && !IAmDisabled)
					{
						MouseInfo mi={0};
						mi.x=LOWORD(msg->lParam); mi.y=HIWORD(msg->lParam);
						SendMouseInfo(&mi);
					}
					else eatit=true;
					break;
				}
				case WM_MOUSEWHEEL:		// Used to pass wheel messages to RISC-OS
				{
					if(tracking)
					{
						MouseInfo mi;
						UINT lines;
						signed short int d=HIWORD(msg->wParam);
						ERRH(SystemParametersInfo(SPI_GETWHEELSCROLLLINES, 0, &lines, 0));
						ConvertMouse(&mi, LOWORD(msg->lParam), HIWORD(msg->lParam), lines*(d/WHEEL_DELTA));
						SendMouseInfo(&mi);
					}
					else eatit=true;
					break;
				}
			}
			if(eatit)
			{	// Make the message as though it never existed
				msg->message=WM_NULL;
				return 0;
			}
		}
	}
	return CallNextHookEx(getmsgh, code, wParam, lParam);
}

void EXPORT MouseSync_Init(int vector)
{
	void *parent, *riscos;
	CallWinPodule_GetRISCOSWindowHandles(&parent, &riscos);
	parentwndh=(HWND) parent;
	mainwndh=(HWND) riscos;
	ROvector=vector;
	ERRH(CallWinPodule_InstallUpcall(RSUpcallHandler, 0));
	ERRH(getmsgh=SetWindowsHookEx(WH_GETMESSAGE, GetMsgIntercept, NULL, GetWindowThreadProcessId(mainwndh, NULL)));
	ERRH(callwndproch=SetWindowsHookEx(WH_CALLWNDPROC, WndProcIntercept, NULL, GetWindowThreadProcessId(mainwndh, NULL)));
	CallWinPodule_SetMouseInputDisable(1);	// Disable motion
}

// Called by RISC-OS in response to the triggering of the vector
void EXPORT MouseSync_GetMouseInfo(MouseInfo *mi)
{
	*mi=mouseinfo;
}

// Called by RISC-OS when mode changes
void EXPORT MouseSync_SetScreenRes(int width, int height, int xeig, int yeig)
{
	void *parent, *riscos;
	CallWinPodule_GetRISCOSWindowHandles(&parent, &riscos);
	parentwndh=(HWND) parent;
	mainwndh=(HWND) riscos;

	scalinginfo.width=(width+1)<<(xeig-1);
	scalinginfo.height=(height+1)<<(yeig-1);
	scalinginfo.xeig=xeig;
	scalinginfo.yeig=xeig; // yeig
	//char temp[256];
	//sprintf(temp,"Scalinginfo width=%d, height=%d\n", scalinginfo.width, scalinginfo.height);
	//OutputDebugString(temp);
	RSClientChange(mainwndh);
}

void EXPORT MouseSync_Deinit()
{
	ERRH(UnhookWindowsHookEx(getmsgh));
	ERRH(UnhookWindowsHookEx(callwndproch));
	CallWinPodule_SetMouseInputDisable(3);
}

#ifdef __cplusplus
}
#endif
